"""Contains the GUI components of the application.

It is implemented as a separate package build on top of the core functionalities.
"""

from . import condetrol, snapshot

__all__ = ["condetrol", "snapshot"]
