from typing import NewType

DeviceParameter = NewType("DeviceParameter", str)
"""A string that represents the name of a device parameter.

The string should have python variable naming conventions.
"""
